<!DOCTYPE html>
<html lang="en-US">
<head>
	{% block head %}
		<meta charset="UTF-8" />
	{% endblock %}
</head>
<body>
	{% block body %}
		<hr>This report was generated by <A href="https://github.com/sandialabs/pecos">Pecos</A>
 		{{ version }}, {{ datestr }}
	{% endblock %}
</body>
</html>